/*
  Copyright 2003 John Knottenbelt
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
*/

#include "unixclientsocket.h"
#include "inputbuffer.h"
#include "outputbuffer.h"
#include "utils.h"

#include <iostream>
#include <cstring>
#include <cstdlib>
#include <cerrno>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <fcntl.h>

UnixClientSocket::UnixClientSocket(const std::string &path)
  : fd(-1), blocking(true), overflowed(false), input(NULL), output(NULL)
{
  connect(path);
}

UnixClientSocket::~UnixClientSocket()
{
  if (fd != -1) {
    close(fd);
    delete input;
    delete output;
  }
}

bool UnixClientSocket::connect(const std::string &path)
{
  fd = socket(PF_UNIX, SOCK_STREAM, 0);

  if (fd < 0) {
    std::cerr << "Fatal error: failed to create socket for connecting:\n"
	      << strerror(errno) << std::endl;
    return false;
  }

  struct sockaddr_un addr;
  addr.sun_family = AF_UNIX;
  strcpy(addr.sun_path, path.c_str());

  if (::connect(fd, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
    close(fd);
    fd = -1;
    return false;
  }

  setNonBlocking(fd);

  input = new InputBuffer(fd, 8192, 32768);
  output = new OutputBuffer(fd, 8192, 32768);
  return true;
}

UnixClientSocket::operator bool() const
{
  return fd >= 0 && !overflowed;
}

void UnixClientSocket::setBlocking(bool mode)
{
  blocking = mode;
}

bool UnixClientSocket::isBlocking() const
{
  return blocking;
}

bool UnixClientSocket::write(const std::string &data)
{
  if (!output->write(data))
    return false;

  if (blocking)
    return output->flush();
  else
    return true;
}

bool UnixClientSocket::getMoreData()
{
  Select s;
  addSelectFDs(s);

  if (!blocking)
    s.addAlarm(0, 0);

  s.wait();

  if (!isReady(s))
    return false;

  if (!processReady(s))
    return false;

  return true;
}

int UnixClientSocket::read(char *data, int count)
{
  if (count > input->getMaxSize())
    return -1;

  if (!blocking) {
  tryAgain:
    if (count < input->getUsed()) 
      return input->read(data, count) ? count : -1;
    else if (input->getUsed() > 0) {
      count = input->getUsed();
      return input->read(data, count) ? count : -1;
    }
    else {
      if (getMoreData())
	goto tryAgain;
      else
	return -1;
    }
  }
  else {
    // Blocking read of N bytes
    while (input->getUsed() < count) {
      if (!getMoreData())
	return -1;
    };

    return input->read(data, count) ? count : - 1;
  }
}

void UnixClientSocket::addSelectFDs(Select &s) const
{
  // Write data out before accepting new data
  if (output->getUsed() > 0)
    output->addSelectFDs(s);
  else
    input->addSelectFDs(s);
}

bool UnixClientSocket::isReady(const Select &s) const
{
  return input->isReady(s) || output->isReady(s);
}

bool UnixClientSocket::processReady(const Select &s) 
{
  if (output->isReady(s)) {
    if (!output->process())
      return false;
  }
  else if (input->isReady(s)) {
    if (!input->process())
      return false;
  }
  return true;
}

bool UnixClientSocket::readLine(std::string &line)
{
  if (blocking) {
    while (true) {
      if (input->readLine(line))
	return true;

      if (input->getFree() == 0)
	return false;
      
      if (!getMoreData())
	return false;
    }
  }
  else 
    return input->readLine(line);
}

bool UnixClientSocket::writeCredentials()
{
  return output->writeCredentials();
}
